/***
 * Juliac
 * Copyright (C) 2009-2013 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt.comp;

import java.io.IOException;
import java.util.List;

import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.loader.Tree;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.conf.GeneratorMap;
import org.objectweb.fractal.juliac.conf.InterceptorSourceCodeGeneratorMap;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.desc.AttributeDesc;
import org.objectweb.fractal.juliac.desc.ComponentDesc;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.proxy.InterceptorClassGenerator;
import org.objectweb.fractal.juliac.proxy.InterceptorSourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.koch.control.interceptor.InterceptorController;
import org.objectweb.fractal.koch.loader.TreeParser;
import org.objectweb.fractal.koch.loader.TreeParserException;

/**
 * A membrane descriptor where the structure of the membrane is component-based.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.2
 */
public class CompBasedMembraneDesc extends MembraneDesc<ComponentDesc> {

	private Juliac jc;
	
    /**
     * Create a new membrane descriptor.
     * 
     * With this membrane descriptor, we should be able to deduce the control
     * interface types (3rd parameter) from the component-based membrane (4th
     * parameter.) Yet it happens that in the COMP mode, we need, in order to
     * generate interceptors, to construct membrane descriptors where the
     * control interface types are either level 0 types (e.g.
     * binding-controller) or meta-level types (e.g. //binding-controller),
     * while keeping the same component-based control membrane.
     * 
     * All this is related to the fact that lifecycle interceptors reference the
     * lifecycle controller implementation classes, and have thus to access the
     * membrane descriptor. If that was not the case, interceptors could be
     * generated based only on the intercepted type.
     * 
     * @param jc            the instance of Juliac this descriptor is attached to
     * @param name          the membrane name (only relevant for component-based membranes)
     * @param descriptor    the Julia membrane descriptor (e.g. primitive)
     * @param ctrlItfTypes  the array of control interface types implemented by this membrane
     * @param ctrlDescs     the structure of the membrane
     * @param jc            the instance of Juliac
     */
    public CompBasedMembraneDesc(
        Juliac jc, String name, String descriptor, InterfaceType[] ctrlItfTypes,
        ComponentDesc ctrlDescs )
	throws IOException {
        
        super(name,descriptor,ctrlItfTypes,ctrlDescs);
        this.jc = jc;
        initInterceptorClassGenerator();
    }
    
    /**
     * Return the implementation class of the control component implementing the
     * specified signature.
     * 
     * @param signature  the signature
     * @return           the class of the control component
     * @throws IllegalArgumentException  if there is no such control component
     * @since 2.2
     */
    public UnifiedClass getCtrlImpl( String signature )
    throws IllegalArgumentException {
        
        UnifiedClass ctrlItf = jc.create(signature);
        
        UnifiedClass uc = getCtrlImpl(ctrlDescs,ctrlItf);
        if( uc == null ) {
            String msg =
                "No control component implementing "+signature+
                " in controller descriptor "+getDescriptor();
            throw new IllegalArgumentException(msg);
        }

        return uc;
    }
    
    /**
     * Return the component descriptor of the interceptor control component
     * contained in the current component descriptor. Return <code>null</code>
     * if there is no such control component.
     * 
     * @since 2.5
     */
    public ComponentDesc getInterceptorControlComponent() {
    	
        /*
         * Iterate until the primitive component is found.
         */
    	ComponentDesc cdesc = ctrlDescs;
    	while( cdesc.getContentClassName() == null ) {
            cdesc = cdesc.getBoundComponent("//"+InterceptorController.NAME);
            if( cdesc == null ) {
                /*
                 * No interceptor controller (e.g. control membrane for
                 * composite components.)
                 */
                return null;
            }
        }
        
        return cdesc;
    }    

    
    // ------------------------------------------------------------------
    // Implementation specific
    // ------------------------------------------------------------------
    
    /**
     * Return the implementation class of the control component implementing the
     * specified control interface. Return <code>null</code> if no control
     * component implements the specified control interface.
     * 
     * @param cdesc    the control component
     * @param ctrlItf  the control interface
     * @return         the class of the control component or <code>null</code>
     */
    private UnifiedClass getCtrlImpl(
        ComponentDesc cdesc, UnifiedClass ctrlItf ) {
        
        /*
         * For component with a content class (primitive), check whether the
         * content class implements the specified control interface.
         */
        String contentClassName = cdesc.getContentClassName();
        if( contentClassName != null ) {
            UnifiedClass uc = jc.create(contentClassName);
            if( ctrlItf.isAssignableFrom(uc) ) {
                return uc;
            }
        }
        
        /*
         * Iterate on sub components.
         */
        List<ComponentDesc> subs = cdesc.getSubComponents();
        for (ComponentDesc sub : subs) {
            UnifiedClass uc = getCtrlImpl(sub,ctrlItf);
            if( uc != null ) {
                return uc;
            }
        }
        
        // No component implements the specified control interface
        return null;
    }
    
    /**
     * Initialize the current interceptor class generator {@link
     * #interceptorClassGenerator} from the current component-based membrane
     * descriptor. Leave it <code>null</code> if there is no interceptor control
     * compoment in the current component-based membrane descriptor.
     * 
     * @since 2.5
     */
    private void initInterceptorClassGenerator() throws IOException {
        
        ComponentDesc cdesc = getInterceptorControlComponent();
        if( cdesc == null ) {
        	return;
        }
        
        AttributeDesc adesc = cdesc.getAttribute("interceptors");
        if( adesc == null ) {
            String msg =
                "Unexpected error: no attribute named 'interceptors' for the "+
                "interceptor controller of the membrane "+
                ctrlDescs;
            throw new JuliacRuntimeException(msg);
        }

        /*
         * Get the configuration string stored in the interceptors property and
         * construct the corresponding Tree.
         */
        String cfg = adesc.getValue();
        TreeParser tp = new TreeParser(cfg);
        Tree t = null;
        try {
            t = tp.parseTree();
        }
        catch( TreeParserException tpe ) {
            IOException ioe = new IOException();
            ioe.initCause(tpe);
            throw ioe;
        }
        
        JuliacConfig jconf = jc.getJuliacConfig();
        GeneratorMap<InterceptorClassGenerator> iclgs =
    		jconf.interceptorClassGenerators();
        InterceptorSourceCodeGeneratorMap iscgmap =
    		jconf.interceptorSourceCodeGenerators();

        // [0]: org.objectweb.fractal.julia.asm.InterceptorClassGenerator
        String[] names = new String[ t.getSize() - 1 ];
        for (int i = 0; i < names.length; i++) {
            Tree tt = t.getSubTree(i+1);
            names[i] = tt.toString();
        }
        boolean mergeable = false;
        InterceptorSourceCodeGeneratorItf[] iscgs =
            iscgmap.get(names,mergeable);
        
        /*
         * Get the interceptor class generator.
         * 
         * If this is a real interceptor class generator, not an association
         * between the name of an ASM bytecode generator or a previously
         * encountered generator, load it in the generator map.
         */
        Tree tt = t.getSubTree(0);
        String name = tt.toString();
        InterceptorClassGenerator icg = iclgs.get(name);
        icg.init(iscgs);
        
        interceptorClassGenerator = icg;
    }
}
