/***
 * Juliac
 * Copyright (C) 2007-2013 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt.comp;

import java.io.IOException;

import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.opt.FCSourceCodeGenerator;
import org.objectweb.fractal.juliac.opt.InitializerClassGenerator;
import org.objectweb.fractal.juliac.proxy.InterceptorClassGenerator;
import org.objectweb.fractal.juliac.proxy.InterceptorSourceCodeGeneratorItf;
import org.objectweb.fractal.koch.control.binding.BindingControllerDef;
import org.objectweb.fractal.koch.control.component.ComponentControllerDef;
import org.objectweb.fractal.koch.control.content.ContentControllerDef;
import org.objectweb.fractal.koch.control.content.SuperControllerDef;
import org.objectweb.fractal.koch.control.lifecycle.LifeCycleControllerDef;
import org.objectweb.fractal.koch.control.name.NameControllerDef;
import org.objectweb.fractal.koch.factory.MCompositeImpl;
import org.objectweb.fractal.koch.factory.MPrimitiveImpl;

/**
 * This class generates the source code associated to Fractal components.
 * The membrane implementation and the initializer implementation are generated.
 * 
 * This generator handles mPrimitive and mComposite components.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public class FCmCompCtrlSourceCodeGenerator
extends FCSourceCodeGenerator<Class<?>> {
    
    // -----------------------------------------------------------------------
    // Implementation of the FCSourceCodeGeneratorItf interface
    // -----------------------------------------------------------------------

    /**
     * Return <code>true</code> if the current source code generator handles the
     * specified controller descriptor.
     * 
     * @param controllerDesc  the controller descriptor to be checked
     * @return                if <code>controllerDesc</code> should be included 
     * @since 2.2
     */
    @Override
    public boolean acceptCtrlDesc( Object controllerDesc ) {
        boolean accept =
            controllerDesc.equals("mPrimitive") ||
            controllerDesc.equals("mComposite");
        return accept;
    }
    
    /**
     * Generate the implementation of the membrane associated to a Fractal
     * component.
     * 
     * @param ct                 the component type
     * @param ctrldesc           the component controller descriptor
     * @param contentClassName   the name of the component content class
     * @return                   data about the membrane implementation
     */
    public MembraneDesc<Class<?>> generateMembraneImpl(
        ComponentType ct, String ctrldesc, String contentClassName )
    throws IOException {
        
        /*
         * Generate business interface implementations.
         */
        InterfaceType[] its = ct.getFcInterfaceTypes();
        generateInterfaceImpl(its,ctrldesc);
        
        /*
         * Generate control interface implementations.
         */
        InterfaceType[] membraneits = getMembraneType(ctrldesc);
        generateInterfaceImpl(membraneits,ctrldesc);
        
        /*
         * Prepare a membrane descriptor.
         */
        Class<?> ctrlImpl = getMembraneClass(ctrldesc);
        MembraneDesc<Class<?>> md =
            new MMembraneDesc(jc,null,ctrldesc,membraneits,ctrlImpl);
        
        /*
         * Generate the lifecycle interceptors associated with control
         * components for all business interfaces of the control components.
         */
        if( ctrldesc.equals("mPrimitive") ) {
            for (InterfaceType it : its) {
                String itname = it.getFcItfName();
                // Exclude control interfaces of control components
                if( ! containsItfName("mPrimitive",itname) ) {
                    // Exclude client interfaces
                    if( ! it.isFcClientItf() ) {
                        InterceptorSourceCodeGeneratorItf lcscg =
                            new LifeCycleSourceCodeGenerator(jc,it,md,false);
                        InterceptorClassGenerator miscg =
                            new InterceptorClassGenerator(
                                jc,
                                new InterceptorSourceCodeGeneratorItf[]{lcscg},
                                it, md, false );
                        jc.generateSourceCode(miscg);
                    }
                }
            }
        }
        
        return md;
    }

    /**
     * Return the source code generator for the initializer class associated
     * with this component source code generator.
     */
    public InitializerClassGenerator<Class<?>> getInitializerClassGenerator() {
        return new InitializermCompCtrlClassGenerator();
    }
    
    
    // -----------------------------------------------------------------------
    // Implementation specific
    // -----------------------------------------------------------------------
    
    /**
     * Return the class implementing the membrane of the specified control
     * component descriptor mPrimitive or mComposite).
     */
    protected Class<?> getMembraneClass( String ctrldesc ) {
        
        if( ctrldesc.equals("mPrimitive") ) {
            return MPrimitiveImpl.class;
        }

        if( ctrldesc.equals("mComposite") ) {
            return MCompositeImpl.class;
        }
        
        String msg = "No such controller desc: "+ctrldesc;
        throw new JuliacRuntimeException(msg);
    }
    
    /**
     * Return the control interface types associated to the specified controller
     * descritptor.
     * 
     * @param ctrldesc  the controller descriptor
     * @return          the associated array of control interface types
     */
    protected InterfaceType[] getMembraneType( String ctrldesc ) {
        
        if( ctrldesc.equals("mPrimitive") ) {
            return
                new InterfaceType[]{
                    ComponentControllerDef.TYPE,
                    BindingControllerDef.TYPE,
                    NameControllerDef.TYPE,
                    SuperControllerDef.TYPE,
                    LifeCycleControllerDef.TYPE
                };
        }
        
        if( ctrldesc.equals("mComposite") ) {
            return
                new InterfaceType[]{
                    ComponentControllerDef.TYPE,
                    BindingControllerDef.TYPE,
                    NameControllerDef.TYPE,
                    SuperControllerDef.TYPE,
                    LifeCycleControllerDef.TYPE,
                    ContentControllerDef.TYPE
                };
        }
        
        String msg = "No such controller desc: "+ctrldesc;
        throw new JuliacRuntimeException(msg);
    }
    
    /**
     * Return <code>true</code> if the specified interface name corresponds to
     * a control interface for the specified controller descriptor.
     * 
     * @param ctrldesc  the controller descriptor
     * @param name      the interface name
     * @since 2.5
     */
    protected boolean containsItfName( String ctrldesc, String name ) {
    	InterfaceType[] its = getMembraneType(ctrldesc);
    	for (InterfaceType it : its) {
			String itname = it.getFcItfName();
			if( itname.equals(name) ) {
				return true;
			}
		}
    	return false;
    }
}
